/* global DeviceManager, CommunicationManager, LibraryManager, utils, LayoutManager, ERRORS, XMRManager, logic, LicenseManager, application, Logger, AppVersionInformation, defined */
(function(){
	"use strict";
	/**
	 * @classdesc Schedule handling.
	 * @type ScheduleManager
	 * @version 1.8.1
	 * @since 1.8.1
	 */
	//jshint -W057
	window.ScheduleManager = new function _ScheduleManager() {

		/* +------------------------+ Private variables +------------------------+ */
		// <editor-fold defaultstate="collapsed" desc="Private variables">
		/**
		 * Flag if collect Interval is running.
		 */
		var v_collectIntervalRunningState = false;
		/**
		 * Flag if the Layout should receive layouts or stay on a specific state like settings screen!
		 */
		var v_runningSchedule = false;

		/**
		 * Object with schedule data.
		 */
		var v_scheduleData = null;

		/**
		 * Reference to collect Interval timer.
		 */
		var v_collectIntervalTimer = null;

		/**
		 * Array that holds which overlay layout is currently showned.
		 */
		var v_CurrentOverlayLayouts = [];

		/**
		 * Flag if the schedule has been run at least once.
		 */
		var v_runScheduleFirstTime = false;
		// </editor-fold>
		

		/**
		 * How many times concurrent connection to XMDS failed.
		 */
		var v_ConcurrentFailedXMDSConnetion = 0;

		/**
		 * Indicate if the content is ready to play.
		 */
		var contentValid = false;

	    /**
		 * Flag if user has shown settings screen.
		 */
	    var v_userShownedSettingsScreen = false;
	    
	    /**
	     * Last "required files" CRC (from CMS).
	     */  
	    var v_localCheckRfCRC = "";
	       
	    /**
	     * Last schedule CRC (from CMS).
	     */  
	    var v_localCheckScheduleCRC = "";
	    
	    // </editor-fold>

		/* +------------------------+ Private functions +------------------------+ */
		// <editor-fold defaultstate="collapsed" desc="Private functions">

		var errorCB = function(cbError){
			logic.nonFatalError(cbError);
		};

		var failedToConnectToCMS = function(){
			v_ConcurrentFailedXMDSConnetion++;
			application.statusPage.failedXMDSConnections(v_ConcurrentFailedXMDSConnetion);
		};

		var resetFailedToConnectToCMS = function(){
			v_ConcurrentFailedXMDSConnetion = 0;
			application.statusPage.failedXMDSConnections(v_ConcurrentFailedXMDSConnetion);
		};



		var requestScheduleFromCMS = function(successCb, failureCb, cmsResponse){
			var createDataForLayoutManager = function(){
				v_collectIntervalRunningState = false;
				if (!v_runScheduleFirstTime){
					v_runScheduleFirstTime = true;
					informLayoutWithSchedule();
				}

				/* As per Dan suggestion we should not send MediaInvetory on collect for bandwidth usage */
				if(successCb){
					successCb();
				}
			};

			var downloadScheduleCompleted = function(){
				contentValid = validateContent();

				if (contentValid){
					/* Content is validated */
					console.debug("Content is valid");
					application.statusPage.scheduleStatus("Up to date");
					ScheduleManager.saveCachedSchedule();
					alwaysSendToCMS(createDataForLayoutManager,function(cbError){
						v_collectIntervalRunningState = false;
						errorCB(cbError);
					});
				}else{
					/* Content is NOT validated */
					console.debug("Content is not valid");
					application.statusPage.scheduleStatus("Content is not valid.");
					v_collectIntervalRunningState = false;
					alwaysSendToCMS(failureCb ,function(cbError){
						v_collectIntervalRunningState = false;
						errorCB(cbError);
					});
				};
				
			};
				
				var downloadSchedule = function(successCb, failureCb){
					
					        /* Only in trial or licensed displays can download schedule */
					
					        console.debug("Download schedule!");
					        // XLG-103 : Add more logs on application start!
					        //application.statusPage.scheduleStatus("Pending...");
					        application.statusPage.scheduleStatus("Download schedule ...");
					
					        var shouldDownload = true;
					        if (cmsResponse.hasOwnProperty('checkSchedule') && cmsResponse.checkSchedule!==""){
					          shouldDownload = cmsResponse.checkSchedule !== v_localCheckScheduleCRC;
					        }
					        if (shouldDownload){
					          CommunicationManager.fetchSchedule(function(data){
					        	v_localCheckScheduleCRC = cmsResponse.checkSchedule;
					            parseSchedule(data);
					            successCb();
					          },failureCb);
					        }else{
					          successCb();
					          application.statusPage.scheduleStatus("Skipped");
					        }
					      };


		

			var parseRequiredFilesCompleted = function(){
				application.statusPage.scheduleStatus("Download schedule ...");
				downloadSchedule(downloadScheduleCompleted,function(cbError){
					v_collectIntervalRunningState = false;
					alwaysSendToCMS(function(){errorCB(cbError);},errorCB);
				});
			};

			var getRequiredFilesCompleted = function(cbObj){

				v_localCheckRfCRC = cmsResponse.checkRf;
				   // XLG-103 : Add more logs on application start!
				    application.statusPage.scheduleStatus("Check required files ...");
				    parseRequiredFiles(
				       parseRequiredFilesCompleted, 
				         function(cbError){
				           v_collectIntervalRunningState = false;
				             alwaysSendToCMS(function(){errorCB(cbError);},errorCB);
				           },
				           cbObj
				     );
			};


			var cbSISuccess = function(){
				if (LicenseManager.inTrial() || LicenseManager.isLicenseValid()){
					console.debug("I have a valid license!");
					/* First option is get Required Files */
					application.statusPage.scheduleStatus("Get required files ...");
	                // Honor download time window
	                var inDownloadWindow=true;
	                if (DeviceManager.Config.getOptions().hasOwnProperty("downloadStartWindow") && 
	                		DeviceManager.Config.getOptions().hasOwnProperty("downloadEndWindow")){
	                  var dlwindow = new Date();
	                  var s = DeviceManager.Config.getOptions().downloadStartWindow;
	                  var e = DeviceManager.Config.getOptions().downloadEndWindow;
	                  try {
	                    inDownloadWindow = dlwindow.inDateTimeRange(s,e);
	                  } catch (e) {
	                    Logger.cmsLog({category:"error", message:e.errorText, type:"error", method:"connectToCMS", thread:"Schedule Manager"});
	                  }
	                }
	                if (inDownloadWindow){
	                	//XLG-154: Bandwidth usage improvements: checkSchedule and checkRf on RegisterDisplay
	                	var shouldDownload = true;
	                	if (cmsResponse.hasOwnProperty('checkRf') && cmsResponse.checkRf!==""){
	                		shouldDownload = cmsResponse.checkRf !== v_localCheckRfCRC;
	                	}
	                	if (shouldDownload){
	                		application.statusPage.requiredFilesSkipped(false);
	                		CommunicationManager.getRequiredFiles(getRequiredFilesCompleted,
	                		function(cbError){
	                	      v_collectIntervalRunningState = false;
	                	        console.error("Error", cbError);
	                	        alwaysSendToCMS(function(){errorCB(cbError);},errorCB);
	                			}
	                		);
	                	}else{
	                		application.statusPage.requiredFilesSkipped(true);
	                		parseRequiredFilesCompleted();
	                  }
	                }else{
	                  console.debug("Outside download window");
	                  v_collectIntervalRunningState = false;
	                  alwaysSendToCMS(function(){
	                    playLocalContent();
	                    if (successCb){
	                      successCb();
	                    }
	                  },errorCB);
	                }
				}else{
					console.debug("Not license, no schedule!");
					v_collectIntervalRunningState = false;
					playLocalContent();
					if (successCb){
						successCb();
					}
				}
			};

			var alwaysSendToCMS = function(successASTCCb, failureASTCCb){
				Logger.submitStats(function(){
					console.debug("Send stats completed");
					Logger.submitLog(function(){
						console.debug("Send log completed");
						if (successASTCCb){
							successASTCCb();
						}
					}, failureASTCCb);
				}, failureASTCCb);
			};

			/* First option */
			CommunicationManager.sendInfo(cbSISuccess,function(cbError){
				v_collectIntervalRunningState = false;
				errorCB(cbError);
			});
		};

		/**
		 * Try connect to the CMS.
		 * @param {Boolean} saveSettings If coming from User parameter view, set it true to save the parameters.
		 * @returns {undefined} None
		 * @version 1.8.1
		 * @since 1.8.1
		 */
		var connectToCMS = function (saveSettings) {
			if (DeviceManager.Config.settingsValid()) {
				if (saveSettings === true)
				{
					DeviceManager.Config.saveConfiguration();
				}

				console.debug("Connect to cms!");
				CommunicationManager.findCMSVersion(
						function(){
							CommunicationManager.ConnectCMS(
									function(successCb, faulureCb, response){
										resetFailedToConnectToCMS();
										connectCMSResponse(null, faulureCb, response);
									}, 
									function(errObj){
										failedToConnectToCMS();
										errorCB(errObj);
									
									},
									{URL:DeviceManager.Config.getURL()}
							);
						},
						function(){
							failedToConnectToCMS();
							console.error("Could not connect to the CMS");
							LayoutManager.toast(null,null,"Could not connect to the CMS");
						}
				);

			} else
			{
				failedToConnectToCMS();
				errorCB(new ERRORS.ErrorWithSettings());
			}
		};


		var connectCMSResponse = function (successCb, failureCb, connectResponse) {
			console.debug("Handle the responce: ", connectResponse);
			v_ConcurrentFailedXMDSConnetion = 0;

			if (connectResponse.hasOwnProperty("status")) /* Normal response from CMS */
			{
				if (connectResponse.status === "0") /* Monitor is licensed! */
				{
					application.statusPage.cmsRegistered(true);
					application.statusPage.cmsRegInfo("Registered.");
					application.statusPage.cmsRegInfoTime(utils.getLocalDate());


					/* Inform that we are a licensed display */
					v_runningSchedule = true;
					/* This is also started from ScheduleManager.start() */
					if (!v_collectIntervalTimer){
						ScheduleManager.startInterval();
					}
					try {
						LayoutManager.startPlaying();
					} catch (e) {
						console.error(e);
					}

					/* TODO: Here we have to connect to XMR. */
					XMRManager.connect();

					if (connectResponse.hasOwnProperty('version_instructions'))
					{
						//console.debug(connectResponse.version_instructions);
						try {
							if (connectResponse.version_instructions !== ""){
								var upg = JSON.parse(connectResponse.version_instructions);

								if (upg.hasOwnProperty('id')){
									if (upg.hasOwnProperty('code')){
										application.statusPage.versionInformationCMS(AppVersionInformation.clientVersion + "." + upg.code);//TODO: Should show the version instead?
									}else{
										application.statusPage.versionInformationCMS(upg.id);
									}
									
									if (localStorage.hasOwnProperty('version_instructions')){
										var myUpg = JSON.parse(localStorage.version_instructions);
	
										if (upg.id !== myUpg.id){
											console.debug("Application needs upgrade!");
											if (AppVersionInformation.isDevVersion){
												LayoutManager.toast(function(){},function(){},"Application needs upgrade (ID missmatch L: "+myUpg.id+" - R:"+upg.id+")!");
											}
											application.applicationUpgrade = true;
											localStorage.version_instructions = connectResponse.version_instructions;
										}
									}else{
										localStorage.version_instructions = connectResponse.version_instructions;
										if (AppVersionInformation.clientCode != upg.code){
											console.debug("Application needs upgrade!");
											if (AppVersionInformation.isDevVersion){
												LayoutManager.toast(function(){},function(){},"Application needs upgrade (No previous ID R: "+upg.id+")!");
											}
											application.applicationUpgrade = true;
										}
									}

								
									if (application.applicationUpgrade && !AppVersionInformation.preventUpdate){
										console.info("Upgrade application!", upg);
										application.statusPage.scheduleStatus("Upgrade application.");
										DeviceManager.Device.upgradeApplication(function(){
											console.debug("Application upgrading!");

											if (AppVersionInformation.isDevVersion){
												setTimeout(function(){
													DeviceManager.Device.powerCommand("reboot");
												}, 15000);
											}else{
												DeviceManager.Device.powerCommand("reboot");
											}
											


										}, function(cbObject){
											Logger.cmsLog({category:"audit", message:cbObject.errorText, type:"info", method:"connectCMSResponse", thread:"Schedule Manager"});
										}, upg.url);
									}
								
								}else{
									Logger.cmsLog({category:"audit", message:'Not compatible version upgrade. ('+connectResponse.version_instructions+')', type:"Audit", method:"connectToCMS", thread:"Schedule Manager"});
								}
							}
						} catch (e) {
							Logger.cmsLog({category:"error", message:e, type:"error", method:"connectToCMS", thread:"Schedule Manager"});
						}
					}

					// Fix for XLG-63 : Request screenshot without working XMR issue
					if (DeviceManager.Config.getOptions().hasOwnProperty("screenShotRequested"))
					{
						if (DeviceManager.Config.getOptions().screenShotRequested === "1")
						{
							ScheduleManager.Commands.takeScreenshot();
						}
					}

					requestScheduleFromCMS(successCb,nextStep, connectResponse);
				}else if (connectResponse.status === "1") /* Monitor added */
				{
					v_runningSchedule = false;
					console.debug("Monitor added!");
					LayoutManager.showSettingsScreen(connectToCMS);
					LayoutManager.toast(null,null,connectResponse.message);
					// Why is this here? Also commented in webOS. 2018-10-23
//					if (failureCb){
//						failureCb();
//					}

				} else if (connectResponse.status === "2") /* Monitor waiting approval */
				{
					v_runningSchedule = false;
					console.debug("Monitor waiting approval!");
					LayoutManager.showSettingsScreen(connectToCMS);
					LayoutManager.toast(null,null,connectResponse.message);
					// Why is this here? Also commented in webOS. 2018-10-23
//					if (failureCb){
//						failureCb();
//					}
				}
			} else if (connectResponse.hasOwnProperty("errorCode")){ /* Error occured connecting to CMS */
				if (connectResponse.errorCode === ERRORS.WRONGSERVERKEYCODE) /* Wrong key */
				{
					v_runningSchedule = false;
					console.debug("Wrong CMS key!");
					LayoutManager.showSettingsScreen(connectToCMS);
					LayoutManager.toast(null,null,connectResponse.errorText);
					if (failureCb){
						failureCb(connectResponse);
					}
				}else{
					console.debug("Connect response error", connectResponse);
					if (failureCb){
						failureCb(connectResponse);
					}
				}
			}else{
				console.warn("CMS responded in unusual fasion.",connectResponse);
				if (failureCb){
					failureCb(connectResponse);
				}
			}
		};

		var collectIntervalLogic = function(successCb, failureCb){
			console.log("Scheduler Collect Interval logic fired!");
			
	        if (v_userShownedSettingsScreen){
	            console.debug("collectInterval:User has opened settings screen! Quit!");
	            return;
	        }
			
			if (v_collectIntervalRunningState){
				console.debug("Scheduler Collect Interval is already running! Quit!");
				return;
			}
			v_collectIntervalRunningState = true;

			var connectFailedCb = function(){
				failedToConnectToCMS();
				if (!playLocalContent()){
					if (failureCb){
						failureCb();
					}
				}
				v_collectIntervalRunningState = false;
			};

			/* Schedule Manager should handle the connection */

			CommunicationManager.ConnectCMS(
					function(s,f,response){
						connectCMSResponse(successCb, failureCb, response);
					},
					connectFailedCb,
					{URL:DeviceManager.Config.getURL()}
			);


		};


		var validateDependents = function(){
			var data;
			try {
	            data = ScheduleManager.getScheduleData().dependents;
	        } catch (e) {
	            return 0;
	        }

			var dependentsValid = 1;
			for (var i = 0, max = data.length; i < max; i++) {
				//jshint -W016
				dependentsValid &= LibraryManager.isFileOk(data[i]);
				//jshint +W016
			}
			console.debug("Dependents are valid:" + dependentsValid);
			return dependentsValid;
		};        

		var validateDefault = function(){
			var data;
			var defaultValid = 1;
			try {
	            data = ScheduleManager.getScheduleData().default;
	        } catch (e) {
	            return 0;
	        }
			if (data.hasOwnProperty("file")){
				//jshint -W016
				defaultValid &= LibraryManager.isFileOk(data.file + ".xlf");
				//jshint +W016
			}else{
				defaultValid = 0;
			}
			if (data.hasOwnProperty("files")){
				for (var i = 0, max = data.files.length; i < max; i++) {
					//jshint -W016
					defaultValid &= LibraryManager.isFileOk(data.files[i]);
					//jshint +W016
				}
			}
			console.debug("Default files are valid:" + defaultValid);
			// application.statusPage.allLayouts(data.file+"* (D)");
			
			if (defaultValid === 1){
		        if (LayoutManager.invalidLayoutIds.indexOf(data.file)===-1){
		            application.statusPage.validLayouts_default(data.file+"* (D)");
		            application.statusPage.invalidLayouts_default("");
	            }else{
	                application.statusPage.validLayouts_default("");
	                application.statusPage.invalidLayouts_default(data.file+"* (D)");
	            }
	        }else{
	            application.statusPage.validLayouts_default("");
	            application.statusPage.invalidLayouts_default(data.file+"* (Default)");
	        }
            application.statusPage.allLayoutsSet.push(
        		application.statusPage.createLayoutInfo(data.file,true,defaultValid,false,false)
    		);
	        data.isValid=defaultValid;
			return defaultValid;
		};

		var validateLayout = function(layoutId)
		{
			var layouts;
			var layout;
			var layoutValid = 1;
			var i, maxi;

			try {
	            layouts = ScheduleManager.getScheduleData().layouts;           
	        } catch (e) {
	            return 0;
	        }
			
			for (i = 0, maxi = layouts.length; i < maxi; i++) {
				if (layouts[i].hasOwnProperty("file")){
					if(layouts[i].file === layoutId.toString()){
						layout = layouts[i];
						break;
					}
				}
			}
			if(!defined(layout)){
				return false;
			}

			//jshint -W016
			layoutValid &= LibraryManager.isFileOk(layout.file + ".xlf");
			//jshint +W016

			for (i = 0, maxi = layout.files.length; i < maxi; i++) {
				//jshint -W016
				layoutValid &= LibraryManager.isFileOk(layout.files[i]);
				//jshint +W016
			}
			var layoutStr = application.statusPage.allLayouts();
			if (layoutStr.length!==0){
				layoutStr+= ", ";
			}
			layoutStr+=layout.file;
			
			console.debug("Layout "+layout.file+" files are valid:" + layoutValid);
			return layoutValid;
		};

		/**
		 * @version 2.0.200
		 * @since 1.8.1
		 */
		var validateLayouts = function()
		{
			var data;
	        
	        try {
	            data = ScheduleManager.getScheduleData().layouts;
	        } catch (e) {
	            return {valid:[], invalid:[]};
	        }

	        var layoutValid = 1;
			var validLayouts = [];
			var invalidLayouts = [];
	        application.statusPage.validLayouts_schedule.removeAll();
	        application.statusPage.invalidLayouts_schedule.removeAll();

			for (var j = 0, jmax = data.length; j < jmax; j++) {
				layoutValid = validateLayout(data[j].file);
				if (layoutValid === 1){
					validLayouts.push(data[j]);
	                if (LayoutManager.invalidLayoutIds.indexOf(data[j].file)===-1){
	                	data[j].isValid = true;
	                	application.statusPage.validLayouts_schedule.push(data[j].file);
	                }else{
	                	application.statusPage.invalidLayouts_schedule.push(data[j].file);
	                }
				}else{
					invalidLayouts.push(data[j]);
					data[j].isValid = false;
	                application.statusPage.invalidLayouts_schedule.push(data[j].file);
				}
				application.statusPage.allLayoutsSet.push(
	        		application.statusPage.createLayoutInfo(data[j].file,false,layoutValid,false,false)
	    		);
			}

			return {valid:validLayouts, invalid:invalidLayouts};
		};        

		var validateOverlay = function()
		{
			var data;

	        try {
	            data = ScheduleManager.getScheduleData().overlays;
	        } catch (e) {
	            return 0;
	        }
	        
			var overlayValid = 1;
	        var curValid = 0;
	        application.statusPage.validLayouts_overlay.removeAll();
	        application.statusPage.invalidLayouts_overlay.removeAll();
			for (var j = 0, jmax = data.length; j < jmax; j++) {
	            curValid = LibraryManager.isFileOk(data[j].file+".xlf");
				//jshint -W016
	            overlayValid &= curValid;
	            //jshint +W016
	            if (curValid){
	              application.statusPage.validLayouts_overlay.push(data[j].file+" (O)");
	            }else{
	              application.statusPage.invalidLayouts_overlay.push(data[j].file+" (O)");
	            }
	            application.statusPage.allLayoutsSet.push(
	        		application.statusPage.createLayoutInfo(data[j].file,false,layoutValid,false,true)
	    		);
	            console.debug("Overlay layout "+data[j].file+" file is valid:" + curValid);
			}
			console.debug("Overlay layouts are valid:" + overlayValid);
			return overlayValid;
		};


		/**
		 * Validates the content
		 */
		var validateContent = function(){
//			scheduleParts['default'] = {};
//			scheduleParts['layouts'] = [];
//			scheduleParts['dependents'] = [];
//			scheduleParts['commands'] = [];
//			scheduleParts['overlays'] = [];

			application.statusPage.allLayoutsSet.removeAll();
			/* First validate dependents */
			var dependentsValid = validateDependents();

			/* Validate default */
			var defaultValid = validateDefault();


			/* Validate layouts */
			var layoutsValid = (validateLayouts().valid.length > 0)*1;

			/* Validate overlays (overlays) */
			var overlayValid = validateOverlay();

	        var ret = dependentsValid & (defaultValid | layoutsValid);// &
																		// overlayValid;

			/* Validate commands (commands) */
			
			return ret;
		};

		var parseRequiredFiles = function(successCb, failureCb, files){
			//WebOSManager.FileSystem.listFiles();
			var options = {};
			options.reqFiles = files;
			options.callback = function(file){

//				/* Inform the media inventory */
				//console.debug("*-* Downloaded:", file);
				LibraryManager.updateFile(file);
			};

			LibraryManager.setRequiredFiles(files);
			application.statusPage.totalFiles(files.length);

			var addFileIdx = 0;

			var downloadMissing = function(successCb, failureCb)  {
				var libFiles = LibraryManager.getLibraryFiles();
				var dlFiles = 0;
				var miInterval = 0;
				var hasFileToDownload = false;

				if (DeviceManager.Config.getOptions().hasOwnProperty("mediaInventoryTimer"))
				{
					miInterval = DeviceManager.Config.getOptions().mediaInventoryTimer * 60000;
				}


				if (miInterval !== 0){
					var v_sendMediaInventoryTimer = setInterval(function(){
						CommunicationManager.sendMediaInventory(function(){
							console.debug("Send Media inventory (timed)!");
						},function(cbSMIObj){
							console.debug("Fail to send Media inventory (timed)!",cbSMIObj);
						});
					}, miInterval);
				}
				var download  = function(successCb, failureCb){
					if (dlFiles<libFiles.length){
						if (!libFiles[dlFiles].isCorrect){
							hasFileToDownload = true;
							libFiles[dlFiles].download(function(){
								// File downloaded
								application.statusPage.downloadedFiles(dlFiles);
								application.statusPage.remainingFiles(libFiles.length - dlFiles);
								console.debug("File "+libFiles[dlFiles].filename+" Downloaded");
								dlFiles++;
								download(successCb, failureCb);
							},function(cbObject){
								//console.debug("File download failed!");
	                            Logger.cmsLog({category:"error", message: "Download failed for file "+libFiles[dlFiles].filename + ". Error:"+JSON.stringify(cbObject), type:"ERROR", method:"download", thread:"parseRequiredFiles"});
								dlFiles++;
								download(successCb, failureCb);
							});
						}else{
							dlFiles++;
							download(successCb, failureCb);
						}
					}else{ // Finish downloading
						// Exit recursive
						if (defined(v_sendMediaInventoryTimer)){ clearInterval(v_sendMediaInventoryTimer); }
						application.statusPage.downloadedFiles(files.length);
						application.statusPage.remainingFiles(0);
						if (hasFileToDownload){
							CommunicationManager.sendMediaInventory(function(){
								console.debug("Send Media inventory (finish)!");
								successCb();
							},function(cbSMIObj){
								console.debug("Fail to send Media inventory (finish)!",cbSMIObj);
								successCb();
							});
						}else{
							successCb();
						}
					}

				};
				CommunicationManager.sendMediaInventory(function(){
					console.debug("Send Media inventory (start)!");
					download(successCb, failureCb);
				},function(cbSMIObj){
					console.debug("Fail to send Media inventory (start)!", cbSMIObj);
					download(successCb, failureCb);
				});
			};

			var addFilesToLib = function(successCb, failureCb, files){

				if (addFileIdx < files.length){
					if (files[addFileIdx].type === "media" ||
							files[addFileIdx].type === "layout" ||
							files[addFileIdx].type === "resource"){
						LibraryManager.addFile(function(){
							addFileIdx++;
							addFilesToLib(successCb, failureCb, files);
						},function(cbObject){
							failureCb(cbObject);
						},files[addFileIdx]);
					}else{
						addFileIdx++;
						addFilesToLib(successCb, failureCb, files);
					}
				}else{
					/* Clean files that are not needed */
					LibraryManager.cleanUpLibrary(function(removed){
						if (removed){
	                        // Remove overlay if exists
	                        var ovl = [];
	                        for (var i = 0, max = v_CurrentOverlayLayouts.length; i < max; i++) {
	                          if (!validateLayout(v_CurrentOverlayLayouts[i].layoutId)){
	                            ovl.push({index:i,layout:v_CurrentOverlayLayouts[i]});
	                          }
	                        }
	                        for (var j = 0, maxj = ovl.length; j < maxj; j++) {
	                            LayoutManager.removeOverlay(ovl[j].layout);
	                            v_CurrentOverlayLayouts.splice(ovl[j].index,1);
	                            //application.statusPage.scheduledLayouts_overlay.remove(ovl[j].layout.layoutId+" (O)");
	                            application.statusPage.filterById(ovl[j].layout.layoutId).forEach(function(itm) {
									itm.isScheduled(false);
								})
	                        }
							var curId = LayoutManager.getCurrentLayoutId();
							if (curId !== "splash"){
								// Check if current layout is valid.
								if (!validateLayout(curId)){
									// Check if default layout is valid.
									if (validateDefault()){
										console.debug("Scheduler: Display default layout!");
										var defLayout = {layoutId:ScheduleManager.getScheduleData()['default'].file,
												scheduleId:"0"};
										LayoutManager.setLayouts([defLayout]);
									}else{
										// play splash
										console.debug("Scheduler: Display splash layout!");
										LayoutManager.setLayouts([]);
	                                    /* Overlay layout not Updating */
	                                    // Remove all overlay layouts
	                                    ScheduleManager.setOverlayLayouts();
									}
								}
							}
						}
						successCb(); /* All files checked */
					}, 
					failureCb);
				}
			};

			addFilesToLib(function(){
				downloadMissing(function(cbObj){
					LibraryManager.saveLibrary();
					successCb(cbObj);
				},failureCb, options
				);
			}, failureCb,files);

		};


		/**
		 * Parse XML document from web service to get the schedule object.
		 * @param {String} data Raw XML Document fom web service.
		 * @return {undefined} undefined
		 */
		var parseSchedule = function(data){
			var i, j, maxi, maxj;
			
			var parser = new DOMParser();
			var xmlDoc = parser.parseFromString(data, "text/xml");

			var reqParentxml = xmlDoc.childNodes[0];

			/* Version 4 */
			/*
        for (var i = 0; i < reqParentxml.childElementCount; i++) {
            var x = reqParentxml.children[i];
            var reqFile = {type: x.localName};
            for (var j = 0; j < x.attributes.length; j++) {
                reqFile[x.attributes[j].name] = x.attributes[j].value;
            }
            reqChilds[i] = reqFile;
        } 
			 */

			/* Version 5 */
			var scheduleParts = parseXMLToObject(reqParentxml);

			scheduleParts.default = {};
			scheduleParts.layouts = [];
			scheduleParts.dependents = [];
			scheduleParts.commands = [];
			scheduleParts.overlays = [];


			var currDatetime = utils.getTimeStamp();
			for (i = 0, maxi = reqParentxml.childElementCount; i < maxi ; i++) {
				var x = reqParentxml.children[i];
				switch (x.nodeName) {
				case "default":
					scheduleParts.default = parseXMLToObject(x);

                    if ((!scheduleParts.default.hasOwnProperty('file')) || 
                            scheduleParts.default.file === "0"){
						scheduleParts.default = {};
						Logger.cmsLog({
						    category:"audit", 
						message: "No default layout set in CMS.",
						method: "parseSchedule",
						thread: "ScheduleManager"
						});
                    }else if (x.children.length>0){
	                	if (x.children[0].children.length>0)
						{
							scheduleParts.default.files = [];
							for (j = 0, maxj = x.children[0].children.length; j < maxj; j++) {
								scheduleParts.default.files.push(x.children[0].children[j].textContent);
							}
						}
					}
					break;
					case "layout":
						var layout = parseXMLToObject(x);
						layout.files = [];
						if (x.children.length>0)
						{
							if (x.children[0].children.length>0)
							{
								for (j = 0, maxj = x.children[0].children.length; j < maxj; j++) {
									layout.files.push(x.children[0].children[j].textContent);
								}
							}
						}
						scheduleParts.layouts.push(layout);
						break;
					case "dependants":
						/* We have a list of files */
						for (j = 0, maxj = x.children.length; j < maxj; j++) {
							//scheduleParts['dependents'].push(parseXMLToObject(x.children[j]));
							scheduleParts.dependents.push(x.children[j].textContent);
						}
						break;
					case "command":
						//<command date=â€‹"2017-04-25 12:â€‹00:â€‹00" scheduleid=â€‹"6" code=â€‹"displayoff">â€‹</command>â€‹
						var command = parseXMLToObject(x);
						command.timeStamp = utils.parseDate(command.date);
						if (currDatetime<command.timeStamp){
							/* Check if command exists. */
							var exists = false;
							for (j = 0, maxj = scheduleParts.commands.length; j < maxj; j++) {
								if (command.date === scheduleParts.commands[j].date){
									if (command.code === scheduleParts.commands[j].code){
										exists = true;
									}
								}
							}
							if (!exists){
								scheduleParts.commands.push(command);
							}
						}
						break;
					case "overlays":
						for (j = 0, maxj = x.children.length; j < maxj; j++) {
							var overlay = parseXMLToObject(x.children[j]);
							overlay.fromTimeStamp = utils.parseDate(overlay.fromdt);
							overlay.toTimeStamp = utils.parseDate(overlay.todt);

							if (currDatetime<overlay.toTimeStamp){
								scheduleParts.overlays.push(overlay);
							}
						}
						break;
					default:
						console.debug(x);
					break;
				}
			}

			v_scheduleData = scheduleParts;
			//return scheduleParts;
		};

		var parseXMLToObject = function(data){
			var obj = {};
			obj.name = data.nodeName;
//			if (data.hasOwnProperty("textContent"))
//			{
//			obj.value = data.nodeName;
//			}

			if (data.attributes.length>0){
				parseXMLAttributesToObjects(obj,data.attributes);
			}else if (data.hasOwnProperty("textContent")){
				if (data.textContent[0] !== '\n'){
					obj.value = data.textContent;
				}
			}

//			if (data.children.length>0){
//			parseXMLChildsToObjects(obj,data.children);
//			}

			return obj;
		};

		var parseXMLAttributesToObjects = function(object, data){
			for (var i = 0, max = data.length; i < max; i++) {
				object[data[i].nodeName] = data[i].value;
			}
		};

		var parseXMLChildsToObjects = function(object, data){
			for (var i = 0, max = data.length; i < max; i++) {
				var obj = {};
				obj.name = data[i].nodeName;
				if (data[i].attributes.length>0){
					parseXMLAttributesToObjects(obj,data[i].attributes);
				}

				if (data[i].children.length>0){
					parseXMLChildsToObjects(obj,data[i].children);
				}

				object[obj.name] = obj;
			}
		};

		/**
		 * Play cached content.
		 * @returns {Boolean} True if played else false.
		 */
		 var playLocalContent = function(){
			 if (ScheduleManager.loadCachedSchedule()){
				 contentValid = validateContent();
				 if (contentValid){
					 console.info("Local schedule is OK. Continue");
					 /* 
					  * That means:
					  * Library is validated.
					  * Schedule is known.
					  */

					 v_runningSchedule = true; // Fix for XLG-39!
					 informLayoutWithSchedule();
					 DeviceManager.Device.collectRealTimeDeviceInfo(null,logic.nonFatalError);
					 return true;
				 }else{
					 console.info("Scheduler: Local schedule is invalid!");
					 if (!v_runningSchedule){
						 console.info("Scheduler: Display slash layout");
						 LayoutManager.setLayouts([]);				 
						 v_runningSchedule = true; // Fix for XLG-39!
						 informLayoutWithSchedule();
						 return true;
				 	 }
				 }
			 }else{
				 console.info("Scheduler: No local schedule!");
				 if (!v_runningSchedule){
					 console.info("Scheduler: Display slash layout");
					 LayoutManager.setLayouts([]);				 
					 v_runningSchedule = true; // Fix for XLG-39!
					 informLayoutWithSchedule();
					 return true;
			 	 }
			 }
			 return false;
		 };

		 var informLayoutWithSchedule = function()
		 {
			 console.debug("Check for schedule changes!");
			 
			 var i, maxi;
			 if (v_scheduleData !== null){
				 if (v_runningSchedule){
					 //If we still download content don't set layouts.
	                 if (!v_collectIntervalRunningState){ 

						 if (contentValid){
							 /* Set Schedule */
							 var layouts = ScheduleManager.getCurrentLayouts();
							 setTimeout(function(){
								 LayoutManager.setLayouts(layouts);
								 /* Set overlay layouts */
								 ScheduleManager.setOverlayLayouts();
							 },0);
						 }else{
							 console.debug("Content not valid to set layouts");
						 }
	                 }

					 /* Run Commands */
					 var currDateTime = utils.getTimeStamp();
					 var commands = ScheduleManager.getScheduleData().commands;
					 for (i = 0, maxi = commands.length; i < maxi; i++) {
						 if ((commands[i].timeStamp-60) <= currDateTime &&
								 currDateTime <= commands[i].timeStamp)
						 {
							 console.debug("Run the command "+commands[i].code+" !");
							 commands[i].delete = true;
							 ScheduleManager.Commands.commandAction(commands[i].code);
						 }else if (currDateTime > commands[i].timeStamp){ /* Command is old. Remove it*/
							 commands[i].delete = true;
						 }
					 }

					 for (i = commands.length-1; i >= 0; --i) {
						 if (commands[i].hasOwnProperty('delete'))
						 {
							 commands.splice(i, 1);
						 }
					 }
				 }
			 }
			 
			 setTimeout(informLayoutWithSchedule,60000);
		 };

		 var nextStep = function(){
			 console.warn("What to do next;");
			 v_collectIntervalRunningState = false;
		 };

		 // </editor-fold>

		 /* +------------------------+ Public functions +------------------------+ */
		 // <editor-fold defaultstate="collapsed" desc="Public functions">

		 this.start = function()
		 {  
			 /**
			  * Callback function if CMS configuration exists.
			  * @param {boolean} cbObj True if exists, false if not.
			  */
			 function cbCmsConfExists(cbObj) {

				 var validSettings = false;
				 if (cbObj === true)
				 {
					 //Load Configuration
					 DeviceManager.Config.loadConfiguration();

					 //Check if it is valid
					 validSettings = DeviceManager.Config.settingsValid();
				 }

				 if (validSettings === false || (!DeviceManager.Config.isCMSValid())) {
					 //Ask user for settings
					 console.debug("logic.showCMSParameterEntry();");
					 LayoutManager.showSettingsScreen(connectToCMS);
				 } else {

					 var options = DeviceManager.Config.getOptions();
					 application.statusPage.cmsAddr(DeviceManager.Config.getURL());
					 application.statusPage.displayName(options.displayName);
					 application.statusPage.screenShotInterval(options.screenShotRequestInterval);
                     application.statusPage.updateWindowStart(options.downloadStartWindow);
                     application.statusPage.updateWindowEnd(options.downloadEndWindow);
                     application.setMenuHideTimeout(options.actionBarDisplayDuration);
					 application.setMenuInitialState(options.actionBarMode);
					 if (options.actionBarMode === "0"){ //Menu is hidden
						 application.hideAppMenu();
					 }else{
						 application.showAppMenu();
					 }


					 if (DeviceManager.Device.getOrientation()!== options.orientation){
						 DeviceManager.Device.setNativePortraitMode(function(){
	                       LayoutManager.orientationChanged();
						 },logic.nonFatalError, options.orientation);
					 }

					 CommunicationManager.init(function () {
						 ScheduleManager.startInterval();
						 LicenseManager.StatusServer.start();
					 }, errorCB);
				 }
			 }

			 DeviceManager.Config.configurationExists(cbCmsConfExists, errorCB);
		 };

		 this.startInterval = function() {
			 /*
			  * Collect interval for requesting staff from the CMS.
			  * @type Number
			  * @sa Default value is 1 hour.
			  */
			 var collectInterval = 60 * 60 * 1000;

			 if (DeviceManager.Config.getOptions().hasOwnProperty('collectInterval'))
			 {
				 collectInterval = DeviceManager.Config.getOptions().collectInterval * 1000;
			 }

			 /* FOR DEVELOPMENT ONLY - CHECK TO CMS INTERVAL */
			 //collectInterval = 1000*60*1;
			 /* FOR DEVELOPMENT ONLY - CHECK TO CMS INTERVAL */

			 /* Set the timer to check to the CMS and update staff! */
			 var intenalCollectIntervalHandler = function(){
				 if (DeviceManager.Config.getOptions().hasOwnProperty('collectInterval'))
				 {
					 collectInterval = DeviceManager.Config.getOptions().collectInterval * 1000;
				 }

				 v_collectIntervalTimer = setTimeout(function(){
					 intenalCollectIntervalHandler();
				 }, collectInterval);
				 console.debug("Fired collect interval timer");
				 collectIntervalLogic();
			 };
			 intenalCollectIntervalHandler();

			 /* Set a new timer to run every minute for the schedule */
			 setTimeout(function(){
				 informLayoutWithSchedule();
			 }, 60000);

		 };

		 /**
		  * Returns schedule data as loaded from remote!
		  * @returns {Object} Schedule data.
		  * @version 1.8.1
		  * @since 1.8.1
		  */
		 this.getScheduleData = function(){
			 return v_scheduleData;
		 };

		 /**
		  * Detects which layouts are scheduled to play and return them as an array.
		  * @returns {Array} Returns array of scheduled layouts.
		  * @see The array is based on the current time.
		  * @version 1.8.1
		  * @since 1.8.1
		  */
		 this.getCurrentLayouts = function(){
			 var i, maxi, prio, l;
			 var schData = ScheduleManager.getScheduleData();
			 var layouts = [];

			 if (schData === null)
			 {
				 application.statusPage.allLayouts("Splash");
				 return layouts;
			 }

			 /* Find the maximum priority of the layouts */
			 var maxprio = Number.MIN_SAFE_INTEGER;
			 
			 var timeValidLayouts = [];
			 
			 /* See if layouts are scheduled to play */
			 var cTime = utils.getTimeStamp();

			 //Timecheck
	         for (var i = 0, max = schData.layouts.length; i < max; i++) {
	        	 if (schData.layouts[i].isValid){
	        		 if ((utils.parseDate(schData.layouts[i].fromdt) < cTime) & (utils.parseDate(schData.layouts[i].todt) > cTime)) /* Layout is valid */
	        		 {
	        			 timeValidLayouts.push(schData.layouts[i]);
	        		 }
	        	 }
	         }
			 
			 
	         //Priority check
	         /* Find the maximum priority of the time valid layouts */
	         for (var i = 0, max = timeValidLayouts.length; i < max; i++) {
	             var prio = parseInt(timeValidLayouts[i].priority);
	             if (prio>maxprio){
	                 maxprio = prio;
	             }
	         }
	         //application.statusPage.scheduledLayouts_schedule.removeAll();
	         application.statusPage.filterByScheduled(true).forEach(function(itm) {
			     itm.isScheduled(false);
			 });
	         for (var j = 0, max = timeValidLayouts.length; j < max; j++) {
	             var prio = parseInt(timeValidLayouts[j].priority);
	             if (prio === maxprio){ /* Play only layouts with the maximum priority */
	                 var l = { layoutId : timeValidLayouts[j].file};
	                 if (timeValidLayouts[j].hasOwnProperty('scheduleid')){
	                     l.scheduleId = timeValidLayouts[j].scheduleid;
	                 }
	                 layouts.push(l);
	                 //application.statusPage.scheduledLayouts_schedule.push(l.layoutId);
	                 //application.statusPage.scheduledLayouts_default("");
	                 application.statusPage.filterById(l.layoutId).forEach(function(itm) {
					     itm.isScheduled(true);
					 })
	             }
	         }
	         
			 if (layouts.length === 0)
			 {
		         if(schData.default.hasOwnProperty('file') && schData.default.file !== "0" ){
					 l = { layoutId : schData.default.file , scheduleId : "0"};
					 layouts.unshift(l);
			         //application.statusPage.scheduledLayouts_default(schData.default.file + "* (D)");
			         application.statusPage.filterById(schData.default.file).forEach(function(itm) {
						itm.isScheduled(true);
					})
		         }
			 }

			 return layouts;
		 };


		 /**
		  * Load schedule from cache if exists.
		  * @returns {Boolean} True if schedule exists else false.
		  * @version 1.8.1
		  * @since 1.8.1
		  */
		 this.loadCachedSchedule = function(){
			 if (localStorage.hasOwnProperty("schedule")){
				 v_scheduleData = JSON.parse(localStorage.schedule);
				 return true;
			 }
			 return false;
		 };

		 /**
		  * Save schedule to cache for offline usage.
		  * @returns None
		  * @version 1.8.1
		  * @since 1.8.1
		  */
		 this.saveCachedSchedule = function(){
			 localStorage.schedule = JSON.stringify(v_scheduleData);
		 };

		 /**
		  * Detects which overlay layouts are scheduled to play inform Layout manager to display them.
		  * @returns None
		  * @see Also remove overlay layouts where their play to time passed.
		  * @version 1.8.1
		  * @since 1.8.1
		  */
		 this.setOverlayLayouts = function(){
			 console.debug("Check Overlay");
			 
	         /* Only if splash screen is not shown */
	         if (LayoutManager.getCurrentLayoutId() === "splash"){

               for (var i = v_CurrentOverlayLayouts.length - 1; i >= 0; i--) {
	             LayoutManager.removeOverlay(v_CurrentOverlayLayouts[i]);
	           }
	           v_CurrentOverlayLayouts = [];
	           //application.statusPage.scheduledLayouts_overlay.removeAll();
	           application.statusPage.filterByOverlay(true).forEach(function(itm) {
	        	   itm.isScheduled(false);
	           });
	           return;
	         }

	        
			 if (v_scheduleData === null){
				 return;
			 }

			 var i, j, maxi;
			 var schData = v_scheduleData.overlays;
			 var layouts = [];

			 /* See if layouts are scheduled to play */
			 var cTime = utils.getTimeStamp();

			 for (i = 0, maxi = schData.length; i < maxi; i++) {
				 if ((schData[i].fromTimeStamp < cTime) && (schData[i].toTimeStamp > cTime)) /* Layout is valid */
				 {
	                 var libfile = LibraryManager.getFileByName(schData[i].file+".xlf");
	                 if (defined(libfile)){
		                 var l = { layoutId : schData[i].file, lastUpdated:libfile.lastChecked};
						 if (schData[i].hasOwnProperty('scheduleid')){
							 l.scheduleId = schData[i].scheduleid;
						 }
						 layouts.push(l);
	                 }
				 }
			 }

			 var cLayoutIdx;
			 /* Remove expired overlay layouts previously showned */
	         /* Also remove overlay layouts that has been updated */
			 for (i = v_CurrentOverlayLayouts.length - 1; i >= 0; i--) {
	             var libfile;
				 cLayoutIdx = -1;
				 for (j = 0; j < layouts.length; j++) {
				   libfile = LibraryManager.getFileByName(schData[i].file+".xlf");
				   if (defined(libfile)){
	                 if (v_CurrentOverlayLayouts[i].lastUpdated !== libfile.lastChecked){
	                     break;
	                 }
					 if(layouts[j].layoutId === v_CurrentOverlayLayouts[i].layoutId){
						 cLayoutIdx = j;
						 break;
					 }
				   }
				 }
	             if (defined(libfile)){
					 if (cLayoutIdx === -1){
						 LayoutManager.removeOverlay(v_CurrentOverlayLayouts[i]);
						 v_CurrentOverlayLayouts.splice(cLayoutIdx,1);
		                 //application.statusPage.scheduledLayouts_overlay.remove(v_CurrentOverlayLayouts[i].layoutId+" (O)");
						 /* Should we remove the old overlay layouts from the schedule data??? */
					 }
	             }
			 }

			 /* Show overlay layouts not previously showned */
			 cLayoutIdx = -1;
			 for (j = 0; j < layouts.length; j++) {
				 for (i = 0, maxi = v_CurrentOverlayLayouts.length; i < maxi; i++) {
					 if(layouts[j].layoutId === v_CurrentOverlayLayouts[i].layoutId){
						 cLayoutIdx = j;
						 break;
					 }
				 }

				 if (cLayoutIdx === -1){
					 LayoutManager.showOverlay(layouts[i]);
					 v_CurrentOverlayLayouts.push(layouts[i]);
	                 //application.statusPage.scheduledLayouts_overlay.push(layouts[i].layoutId+" (O)");
				 }
			 }
			 
	         //Display overlay layouts to status page
	         //application.statusPage.scheduledLayouts_overlay.removeAll();
			 application.statusPage.filterByOverlay(true).forEach(function(itm) {
			     itm.isScheduled(false);
			 });
	         for (var l = 0; l < layouts.length; l++) {
	           //application.statusPage.scheduledLayouts_overlay.push(layouts[l].layoutId+" (O)");
        	     application.statusPage.filterById(layouts[l].layoutId).forEach(function(itm) {
				     itm.isScheduled(true);
				 })
	         }
		 };

		/**
		 * Try connect to the CMS.
		 * @param {Boolean} saveSettings If coming from User parameter view, set it true to save the parameters.
		 * @returns None
		 * @version 1.8.1
		 * @since 1.8.1
		 */
		this.ConnectToCMS = function(saveSettings)
		{
			//Clear the library and schedule here.
			LibraryManager.clearAllFiles();
			v_runningSchedule = false;
			v_scheduleData = null;
			LayoutManager.stopPlaying();
	        v_userShownedSettingsScreen = false;
			//Also request a new license.
			LicenseManager.checkLicenseServer(function(){
				//console.log("Check license...");
			},function(errObj){
				console.log("Error checking license...",errObj);
			});
			connectToCMS(saveSettings);
		};

		/**
		 * Clear which overlay layout is shown. It is only used when displaying settings screen (app.js).
		 * @returns None
		 * @version 1.8.1
		 * @since 1.8.1
		 */
	    this.userOpenSettingsScreen = function(){
			v_CurrentOverlayLayouts = [];
			v_runningSchedule = false;
			v_userShownedSettingsScreen = true;
		};
		    
		    
		/**
		 * Make the Schedule Manager to revalidate the content. Usually after a media failed to play
		 * @returns None
		 * @version 1.8.1
		 * @since 1.8.1
		 */
		this.revalidateContent = function(){
			validateContent();

		};

		// <editor-fold defaultstate="collapsed" desc="Commands">
		this.Commands = new function _Commands(){

			/**
			 * Layouts that are shown by changeLayout command.
			 */
			var changeLayoutLayouts = [];

			/**
			 * Holds schedule layouts from changeLayout
			 */
			var stoppedScheduleLayouts = [];

			/**
			 * Timer for the duration of the changed layout.
			 */
			var changeLayoutTimer;

	        /**
	         * Take a screenshot. It will also send runtime information to CMS
	         * @returns {undefined}
	         * @version 1.8.1
	         * @since 1.8.1
	         */
			this.takeScreenshot = function(){
				var attempts = 0;

				var cbFailure = function(cbObject){
					console.error("Request screenshot error", cbObject);
				};
				/*
	            var bd = document.getElementsByTagName("*");
	            for (var i = 0, max = bd.length; i < max; i++) {
	                        bd[i].classList.add("paused");
	            }
	            var ve = document.getElementsByTagName("video");
	            for (var i = 0, max = ve.length; i < max; i++) {
	                ve[i].pause();
	            }
				 */
				var intTakeScreenshot = function(){
					var ssopt;
					var opts = DeviceManager.Config.getOptions();
					if (opts.hasOwnProperty("screenShotSize")){
						ssopt = opts.screenShotSize;
					}
					DeviceManager.Device.screenShot(
						function(cbObject){
							console.debug("Screen capture succeded");
							var obj = { imagedata: cbObject.data };
							logic.status.setLastCommandStatus("true");
							CommunicationManager.sendScreenshot(function(){
								console.debug("Screenshot send!");
								CommunicationManager.sendInfo(null,cbFailure);
							},cbFailure,obj);
						},function(cbOject){
							if (attempts<100){
								console.warn("Screenshot attemp "+attempts+"/100!");
								attempts++;
								setTimeout(intTakeScreenshot,1000);
							}else{
								cbFailure(cbOject);
							}
						},
						ssopt
					 );
				 };
				 intTakeScreenshot();

				 //Also send the status window
				 CommunicationManager.sendStatusInfo(function(){},function(){});
			 };

			 /**
			  * Change the layout.
			  * @param {Number} layoutId The layout Id.
			  * @param {Number} duration The duration to play. 0 means default which is 5.
			  * @param {Boolean} downloadRequired Has to download the resource.
			  * @param { "replace" | "queue" } changeMode The way to change the schedule.<br>
			  * Replace will immidiately change the layout.<br>
			  * Queue will change the layout after the current layout finish playing.
			  * @returns None
			  */
			 this.changeLayout = function(layoutId, duration, downloadRequired, changeMode){
				 if (contentValid){
					 var f_changeLayout = function(){
						 v_runningSchedule = false;
						 console.debug("Command:Change layout to "+layoutId+" with duration " +duration+ ". Change mode is "+changeMode+". Download required : "+downloadRequired);

						 var durationExpired = function(){
							 changeLayoutTimer = setTimeout(function(){
								 changeLayoutLayouts.shift();
								 if (changeLayoutLayouts.length>0){
									 console.debug("Command:Change to layout "+changeLayoutLayouts[0].layoutId+"!");
									 LayoutManager.setLayouts(changeLayoutLayouts[0].layoutId+"");
									 durationExpired();
								 }else{
									 console.debug("Command:Change back to normal schedule!");
									 v_runningSchedule = true;
									 LayoutManager.setLayouts(stoppedScheduleLayouts);
								 }
							 }, changeLayoutLayouts[0].duration*1000);
						 };

						 var dryRun = false;
						 if (changeLayoutLayouts.length === 0){
							 stoppedScheduleLayouts = ScheduleManager.getCurrentLayouts();
							 dryRun = true;
						 }

						 if (duration*1 === 0){
							 duration = 5;
						 }
						 var newlayout = {layoutId:layoutId, duration:duration};

						 if (changeMode === "replace"){
							 clearTimeout(changeLayoutTimer);
							 changeLayoutLayouts = [];
							 dryRun = true;
						 }
						 changeLayoutLayouts.push(newlayout);

						 if (dryRun){
							 LayoutManager.setLayouts(changeLayoutLayouts[0].layoutId+"");
							 durationExpired();
						 }
					 };

					 if (downloadRequired===true){
						 collectIntervalLogic(function(){
							 f_changeLayout();
						 },function(cbError){
							 logic.nonFatalError(cbError);
						 }); /* This has to end before change the layout but it is async. What to do??? */
					 }else{
						 f_changeLayout();
					 }
				 }else{
					 console.debug("Content not valid to run command changeLayout!");
				 }
			 };

			 /**
			  * Run collect now.
			  */
			 this.collectNow = function(){
				 console.debug("Run collect!");
				 collectIntervalLogic();
			 };

			 this.commandAction = function(code){

				 var execute = function(cmdString){
					 switch (cmdString) {
					 //<editor-fold defaultstate="collapsed" desc="Power Commands">
					 case "reboot":
						 DeviceManager.Device.powerCommand(DeviceManager.Device.Power.REBOOT);
						 break;
					 case "poweroff":
						 DeviceManager.Device.powerCommand(DeviceManager.Device.Power.SHUTDOWN);
						 break;
					 case "displayoff":
						 DeviceManager.Device.setDisplayMode(
								 function(){
									 logic.status.setLastCommandStatus("true");
									 CommunicationManager.sendInfo(null,logic.nonFatalError);
								 },
								 function(cbError){
									 logic.nonFatalError(cbError);
								 },
								 DeviceManager.Device.Power.DISPLAY_OFF
						 );
						 break;
					 case "displayon":
						 DeviceManager.Device.setDisplayMode(
								 function(){
									 logic.status.setLastCommandStatus("true");
									 CommunicationManager.sendInfo(null,logic.nonFatalError);
								 },
								 function(cbError){
									 logic.nonFatalError(cbError);
								 },
								 DeviceManager.Device.Power.DISPLAY_ON
						 );
						 break;
						 //</editor-fold>
					 default:
						 var cmds = DeviceManager.Config.getCommands();
					 if (cmds.hasOwnProperty(code)){
						 logic.status.setLastCommandStatus("false");
						 console.debug("Command "+code+" found!", 
								 {method:"commandAction", execute:cmds[code].commandString, validate:cmds[code].validationString} );
					 }else{
						 logic.status.setLastCommandStatus("false");
						 console.warn("requestCommandAction "+code+" not understood!!!");                
					 }
					 break;
					 }  
				 };

				 var cmds = DeviceManager.Config.getCommands();
				 if (cmds.hasOwnProperty(code)){
					 console.debug("Command "+code+" found!", 
							 {method:"commandAction", execute:cmds[code].commandString, validate:cmds[code].validationString});
					 execute(cmds[code].commandString,cmds[code].validationString);

				 }else{
					 logic.status.setLastCommandStatus("false");
					 console.warn("requestCommandAction "+code+" not understood!!!");                
				 }


			 };

			 /**
			  * Show overlay layout.
			  * @param {Number} layoutId The layout Id.
			  * @param {Number} duration The duration to play. 0 means default which is 5.
			  * @param {Boolean} downloadRequired Has to download the resource.
			  * @returns None
			  * @see It will not show again a scheduled overlay layout.
			  */
			 this.overlayLayout = function(layoutId, duration, downloadRequired)
			 {
				 var f_overlayLayout = function()
				 {
					 var scheduledOverlays = ScheduleManager.getScheduleData().overlays;
					 for (var i = 0, max = scheduledOverlays.length; i < max; i++) {
						 if (scheduledOverlays[i].file === layoutId){
							 console.debug("Overlay layout " + layoutId + " is already shown. Don't show again!");
							 return;
						 }
					 }
					 console.debug("Show overlay layout " + layoutId + " for "+ duration + " seconds.");
					 if (duration*1 === 0){
						 duration = 5;
					 }
					 LayoutManager.showOverlay([{layoutId:layoutId+"", scheduleId : "0"}]);

					 setTimeout(function(){
						 console.debug("Hide overlay layout " + layoutId + ".");
						 LayoutManager.removeOverlay([{layoutId:layoutId+"", scheduleId : "0"}]);
					 }, duration*1000);
				 };

				 if (downloadRequired===true){
					 collectIntervalLogic(function(){
						 f_overlayLayout();
					 },function(cbError){
						 logic.nonFatalError(cbError);
					 });
				 }else{
					 f_overlayLayout();
				 }    
			 };

			 this.rekeyAction = function(){
				 var st = utils.getTimeStamp();
				 console.info("RekeyAction started");
				 XMRManager.createKey();
				 var et = utils.getTimeStamp();
				 var tt = et - st;
				 console.info("RekeyAction ended in " + tt + " seconds.");
				 //XMRManager.

				 connectToCMS(true);
			 };

			 this.revertToSchedule = function(){
				 console.info("Revert To Schedule");
				 v_runningSchedule = true;
				 LayoutManager.setLayouts(stoppedScheduleLayouts);
				 changeLayoutLayouts = [];
				 changeLayoutTimer = null;
			 };
		 }();
		 // </editor-fold desc="Commands">


		// </editor-fold desc="Public functions">
	}();
}());